;;; misc-ustils.el --- Strip CHARS from STRING and more

;; Version: 1.1

;; Author: Andreas Roehler <andreas.roehler@easy-emacs.de>, unless indicated otherwise 

;; Keywords: convenience


;; This file is free software; you can redistribute it

;; and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2, or (at your option) any later version.

;; This file is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GNU Emacs; see the file COPYING.  If not, write to
;; the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
;; Boston, MA 02110-1301, USA.

;;; Commentary: Provide some functions for convenience

;; Changes to previous version:

;;; History:
;; 

;;; Code:

(defcustom strip-chars-before  " \t\r\n\f*"
 "Regexp indicating which chars shall be stripped before STRING - which is defined by `string-chars-preserve'."

:type 'string
:group 'convenience)

;; (setq strip-chars-after  " \t\r\n\f*")
(defcustom strip-chars-after  " \t\r\n\f*\\'"
  "Regexp indicating which chars shall be stripped after STRING - which is defined by `string-chars-preserve'."

:type 'string
:group 'convenience)

(defcustom string-chars-preserve "\\(.*?\\)"
  "Chars preserved of STRING.
`strip-chars-after' and
`strip-chars-before' indicate what class of chars to strip."
  :type 'string
  :group 'convenience)

;; This string-strip results from a discussion at
;; http://lists.gnu.org/archive/html/emacs-devel/2006-06/msg00428.html
;; Thanks all contributors.
(defun string-strip (str &optional chars-before chars-after chars-preserve)
  "Return a copy of STR, CHARS removed, like
\(string-strip \" foo   \") -> \"foo\" 
Spezify chars to remove via
`CHARS-BEFORE' and `CHARS-AFTER' default is \"[ \\t\\r\\n\\f]*\",
i.e. spaces, tabs, carriage returns, newlines and newpages.
`CHARS-PRESERVE' must be a parentized expression,
it defaults to \"\\(.*?\\)\"
"
  (let ((s-c-b (or chars-before
		   strip-chars-before))
	(s-c-a (or chars-after
		   strip-chars-after))
	(s-c-p (or chars-preserve
		   string-chars-preserve)))
    (string-match
     (concat "\\`[" s-c-b"]*" s-c-p "[" s-c-a "]*\\'") str)
    (match-string 1 str)))



;;;###autoload
(unless (functionp 'in-string-p)
  (defun in-string-p (&optional pos)
    (let ((orig (or pos (point))))
      (save-excursion
        (save-restriction
          (widen)
          (beginning-of-defun)
          (numberp (nth 3 (parse-partial-sexp (point) orig))))))))

(unless (functionp 'empty-stringp)
  (defun empty-stringp (string)
    (string= "" string)))

(unless (boundp 'empty-line-p-chars)
  (defcustom empty-line-p-chars "^[ \t\f\r]*$"
    "Empty-line-p-chars."
    :type 'regexp
    :group 'convenience))

(unless (functionp 'empty-line-p) 
  (defun empty-line-p (&optional iact)
    "Returns t if cursor is at an empty line, nil otherwise."
    (interactive "p")
    (save-excursion
      (beginning-of-line)
      (when iact
        (message "%s" (looking-at empty-line-p-chars)))
      (looking-at empty-line-p-chars))))

;; redefining it here as existing function returs false positives
;; with GNU Emacs 23.1.50.1 (i686-pc-linux-gnu, GTK+ Version 2.12.0) of 2009-07-25 
;; (unless (featurep 'xemacs)
;; ;; (unless (functionp 'region-exists-p)
;; (defun region-exists-p ()
;; "and mark-active transient-mark-mode
;; (not (eq (region-beginning) (region-end)"
;; (and mark-active transient-mark-mode
;; (not (eq (condition-case nil (region-beginning)(error nil))  (condition-case nil (region-end) (error nil)))))))

(unless (functionp 'indent-with-spaces)
  (defun indent-with-spaces (column)
    "Indent with spaces to column "
    (interactive "*")
    (while (< (current-column) column)
      (insert " "))
    (forward-char 1)))

(when (featurep 'xemacs) 
  (unless (functionp 'looking-back)
    ;; from GNU Emacs subr.el
    (defun looking-back (regexp &optional limit greedy)
      "Return non-nil if text before point matches regular expression REGEXP.
    Like `looking-at' except matches before point, and is slower.
    LIMIT if non-nil speeds up the search by specifying a minimum
    starting position, to avoid checking matches that would start
    before LIMIT.
    If GREEDY is non-nil, extend the match backwards as far as possible,
    stopping when a single additional previous character cannot be part
    of a match for REGEXP."
      (let ((start (point))
            (pos
             (save-excursion
               (and (re-search-backward (concat "\\(?:" regexp "\\)\\=") limit t)
                    (point)))))
        (if (and greedy pos)
            (save-restriction
              (narrow-to-region (point-min) start)
              (while (and (> pos (point-min))
                          (save-excursion
                            (goto-char pos)
                            (backward-char 1)
                            (looking-at (concat "\\(?:"  regexp "\\)\\'"))))
                (setq pos (1- pos)))
              (save-excursion
                (goto-char pos)
                (looking-at (concat "\\(?:"  regexp "\\)\\'")))))
        (not (null pos))))))

(when (eq 0 (string-match "^23" emacs-version))
  (defalias 'region-exists-p 'use-region-p))

(provide 'misc-utils)
;;; misc-utils.el ends here
