
;;; thingatpt-utils-base.el --- th-at-point edit functions

;; Thingatpt-utils Version: 1.13

;; Copyright (C) 2006, 2007, 2008, 2009 Andreas Roehler, unless
;; indicated otherwise

;; Author: Andreas Roehler <andreas.roehler@easy-emacs.de>, unless
;; indicated otherwise

;; This file is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; This file is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.
;; You should have received a copy of the GNU General Public License
;; along with GNU Emacs; see the file COPYING.  If not, write to
;; the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
;; Boston, MA 02110-1301, USA.

;; Keywords: convenience

;;; Major changes to previous version:
;; New class of functions hiding things in region, useful when
;; editing large regexps, providing better readability.

;; For example if a buffer contains something like

;; ("B\[ \t\r\n\f\]?r\[ \t\r\n\f\]?a\[ \t\r\n\f\]?n\[ \t\r\n\f\]?d\[ \t\r\n\f\]?e\[ \t\r\n\f\]?.\[ \t\r\n\f\]?b\[ \t\r\n\f\]?u\[ \t\r\n\f\]?r\[ \t\r\n\f\]?g" . "Brandenburg")

;; after `M-x hide-bracketed-atpt' its shown as

;; ("B?r?a?n?d?e?.?b?u?r?g" . "Brandenburg")

;; a second `M-x hide-bracketed-atpt' removes hiding


;;; Commentary:

;; A set of functions to return, mover over or
;; manipulate a given THING. THING may be a well known
;; form as symbol', `list', `sexp', `defun' but also a
;; new defined and abstract thing. 

;; Call one of the test-functions `th-delimtest', 
;; `th-mv-test', `th-test' with come chars in
;; scratch-buffer or any else changable buffer to get
;; an impression

;; The idea comes from Mike Williams
;; <mikew@gopher.dosli.govt.nz>, author of
;; thingatpt.el

;; The goal is to have a set of similar forms. For
;; example, to provide a word with double-quotes around
;; it, call doublequote-word-atpt. In a similar way you
;; may double-quote not just a word, but any object
;; instrumented here as THING. To make parentheses
;; around it call parentize-word-atpt, etc.

;; Move-functions of this package differ from common
;; behaviour in such, as `forward-word-atpt' stops
;; not after THING, but on the last char of
;; THING. That's in order to enable a call of
;; thing-at-point functions at the end
;; position. Otherwise, when cursor stops after word
;; (THING) as with known `forward-word', word-atpt would return
;; nil. 

;; To see other features, maybe try separate-list-atpt
;; or comment-list-atpt while point is inside a
;; list. Try it again with an abstract char-class as
;; [:alnum:], i.e. try comment-alnum-atpt,
;; brace-alnum-atpt etc.

;; Call `list-of-things-atpt' to see which objects are
;; presently instrumented, beside `delimited-atpt', where
;; selection specified on the fly (because of the wide range of
;; chars, designing such a possible THING).

;; Most THINGS are provided with a set of functions at,
;; before and after point - i.e. to call with
;; ACTION-THING-atpt, -bf, af.

;; This utility comes with test-functions which return
;; the possible results of most functions (exception
;; are the kill-fns). Call th-test, th-mv-test
;; or th-delimtest over text. That-delimtest
;; changes but restores the buffer. Customize the speed
;; of execution via `th-delimtest-delay' and
;; `th-mv-test-delay.'

;; Diffs to basics of required thingatpt.el:
;; `bounds-of-thing-at-point' is replaced by a new
;; `th-bounds', which now first searches
;; backward. As a consequence several
;; `beginning-op-at' and `end-op-at' constructs had
;; to be rewritten.

;; Behavior in general is not validating; i.e. if you
;; call url-atpt and there is no url, all chars at
;; point may be picked, which could be part of a
;; url. Sometimes, however, a kind of validation may be
;; introduced.

;; If calling from a program `bounds-of-THING-atpt' is
;; recommended as an entry-point. It delivers a list
;; with beg and end positions.  

;; In case of trouble, please send me a bug report. Any
;; ideas and comments welcome.

;; You might be interested also to visit Drew Adam's
;; http://www.emacswiki.org/emacs/thingatpt+.el
;; which predates this approach and was helpful writing it.

;; Thing-at-point delivers a portion of the
;; buffer. Thats useful, if THING is not as easy to grasp as a word.
;; For example the first string of an objekt like:

;; ("4[[:punct:] \t\r\n\f]? [[:punct:] \t\r\n\f]?C[[:punct:] \t\r\n\f]?.[[:punct:] \t\r\n\f]?2[[:punct:] \t\r\n\f]?4[[:punct:] \t\r\n\f]?6[[:punct:] \t\r\n\f]?4[[:punct:] \t\r\n\f]?/[[:punct:] \t\r\n\f]?0[[:punct:] \t\r\n\f]?3[[:punct:] \t\r\n\f]? [[:punct:] \t\r\n\f]?B" . "blah blub B")

;; Put the cursor somewhere into the first string:
;; `doublequoted-atpt' will return it, copied into the kill-ring,
;; enabling yanking it and a lot of further actions.

;; `doublequoted-atpt' here is to
;; (global-set-key [(super \")] 'doublequoted-atpt)

;; alike a range of similar commands exist:
;; (global-set-key [(super \')] 'singlequoted-atpt)
;; (global-set-key [(super \))] 'parentized-atpt)
;; (global-set-key [(super \/)] 'slashed-atpt)
;; (global-set-key [(super \\)] 'backslashed-atpt)
;; (global-set-key [(super \])] 'bracketed-atpt)
;; (global-set-key [(super \})] 'braced-atpt)

;; So far THING is simply picked up.

;; Different approach combines copying, deleting with delimiting

;; if region is active:

;; (global-set-key [(control c) (\")] 'doublequote-or-copy-atpt)

;; will provide doublequotes at beginning and end of region.

;; With negative argument it deletes the doublequoted portion under
;; point.

;; Without any argument these functions return as their simplier
;; counterparts

;; With universal argument [(control u)] delimiters --i.e. doublequotes, slashes, whatever-- are stripped.

;; THING as a buffer substring is determined by
;; move-functions specified for thingatpt, called
;; beginning-op-at and end-op-at. Point is stored
;; after move, beginning and end delivered as pair: as
;; consed bounds-of-thing. It's easy to write your own
;; thing-at-point functions that way. You need the
;; caller and both move forms:

;; (defun MY-FORM-atpt (&optional arg) 
;;   " "
;;   (interactive "p")
;;   (th 'MY-FORM arg))

;; (put 'MY-FORM 'beginning-op-at 
;;            (lambda () MY-FORWARD-MOVE-CODE))

;; (put 'MY-FORM 'end-op-at
;;      (lambda () MY-BACKWARD-MOVE-CODE))

;; For example if you want to pick all chars at point
;; which are written between a string "AAA" and a
;; "BBB", which may exist as
;; AAA Luckily detected a lot of things! BBB
;; After evaluation of
;; (put 'MY-FORM 'beginning-op-at
;;      (lambda ()
;;        (search-backward "AAA" nil t 1)
;;        ;; step chars of search expression back
;;        (forward-char 3)))
;; 
;; (put 'MY-FORM 'end-op-at
;;      (lambda ()
;;        (search-forward "BBB" nil t 1)
;;        (forward-char -3)))
;; together with the functions definition above, it's ready.
;; M-x MY-FORM-atpt 
;; (while point inside) you should see: 
;; " Luckily detected a lot of things! " 
;; in the minibuffer.

;; Some keys 

;; (define-key emacs-lisp-mode-map [(control c)(q)] 'list-forward-atpt)
;; (define-key emacs-lisp-mode-map [(super c)())] 'symbol-parentize-atpt)
;; (define-key emacs-lisp-mode-map [(super c)(n)] 'region-parentize-atpt)
;; (global-set-key [(control c) (")] 'doublequoted-or-copy-atpt)
;; (global-set-key [(control c) (')] 'singlequoted-or-copy-atpt)
;; (global-set-key [(control c) (})] 'braced-or-copy-atpt)
;; (global-set-key [(control c)(<)] 'angled-greater-or-copy-atpt)
;; (global-set-key [(control c)(>)] 'angled-lower-or-copy-atpt)
;; (global-set-key [(control c)(()] 'paren-atpt)
;; (global-set-key [(control c)())] 'parentized-or-copy-atpt)
;; (global-set-key [(control c)(/)] 'slashed-or-copy-atpt)
;; (global-set-key [(control c)(\)] 'backslashed-or-copy-atpt)
;; (global-set-key [(control c)(])] 'bracketed-or-copy-atpt)
;; (global-set-key [(control c)(i)] 'parentized-atpt)
;; (global-set-key [(control c)(k)] 'list-kill-atpt)
;; (global-set-key [(control c)(m)] 'mark-word-atpt)
;; (global-set-key [(control c)(o)] 'comment-atpt)
;; (global-set-key [(control c)(p)] 'graph-parentize-atpt)
;; (global-set-key [(f11)] 'list-separate-atpt)
;; (global-set-key [(meta f5)] 'syntax-atpt)
;; (global-set-key [(super ")] 'doublequoted-atpt)
;; (global-set-key [(super ')] 'singlequoted-atpt)
;; (global-set-key [(super ))] 'parentized-atpt)
;; (global-set-key [(super /)] 'slashed-atpt)
;; (global-set-key [(super <)] 'angled-greater-atpt)
;; (global-set-key [(super >)] 'angled-lower-atpt)
;; (global-set-key [(super \)] 'backslashed-atpt)
;; (global-set-key [(super ])] 'bracketed-atpt)
;; (global-set-key [(super })] 'braced-atpt)
;; (global-set-key [(super c)(b)] 'word-brace-atpt)
;; (global-set-key [(super d)] 'delimited-atpt)
;; (global-set-key [(super g)] 'doublequote-graph-atpt)
;; (global-set-key [(super m)] 'symbol-atpt)
;; (global-set-key [(super o)] 'word-atpt)
;; (global-set-key [(super p)] 'copy-paragraph-atpt)
;; (global-set-key [(super u)] 'url-atpt)
;; (global-set-key [(super y)] 'graph-atpt)



;;; Code:


(require 'string-strip)
(require 'beg-end)
(require 'sh-beg-end)
(require 'misc-utils)

(when (featurep 'xemacs) (require 'overlay))

(defcustom thing-mark-region t
  "If a region should be activated for THING. "
  :type 'boolean
:group 'convenience
)

;; (defvar th-orig 0
;; "Correct orig according to delimiter-length")

(if (featurep 'xemacs) 
    (setq thatpt-directory "~/xemacs-thingatpt")
  (setq thatpt-directory "~/emacs-thingatpt"))

(make-directory thatpt-directory t)

(when (featurep 'xemacs)
  (defcustom alnum "\\sw"
  "Rexexp to specify the character class 
Follows word-syntax. Use something like
   \"[a-zA-ZäöüßÄÖÜ0-9]\" maybe instead.
`unibyte' and  `multibyte' class is unused i.e. set to \".\""
  :type 'regexp
  :group 'convenience))

(when (featurep 'xemacs)
  (defcustom alpha "[a-zA-ZäöüßÄÖÜ]"
  "Rexexp to specify the character class 
Change it, if you want to pick strings differently.
XEmacs-users: `unibyte' and  `multibyte' class is unused i.e. set to \".\""
  :type 'regexp
  :group 'convenience))

(when (featurep 'xemacs)
  (defcustom ascii "[\000-\177]"
    "Rexexp to specify the character class 
Change it, if you want to pick strings differently.
XEmacs-users: `unibyte' and `multibyte' class is unused i.e. set to \".\""
    :type 'regexp
    :group 'convenience))

(when (featurep 'xemacs)
  (defcustom blank "[ \t]"
    "Rexexp to specify the character class 
Change it, if you want to pick strings differently.
XEmacs-users: `unibyte' and `multibyte' class is unused i.e. set to \".\""
    :type 'regexp
    :group 'convenience))

(when (featurep 'xemacs)
  (defcustom cntrl "[\000-\006]\016-\037]"
    "Rexexp to specify the character class 
Change it, if you want to pick strings differently.
XEmacs-users: `unibyte' and `multibyte' class is unused i.e. set to \".\""
    :type 'regexp
    :group 'convenience))

(when (featurep 'xemacs)
  (defcustom digit "[0-9]"
  "Rexexp to specify the character class 
Change it, if you want to pick strings differently.
XEmacs-users: `unibyte' and  `multibyte' class is unused i.e. set to \".\""
  :type 'regexp
  :group 'convenience))

(when (featurep 'xemacs)
  (defcustom graph "[\041-\177\241-\377]"
  "Rexexp to specify the character class 
Change it, if you want to pick strings differently.
XEmacs-users: `unibyte' and  `multibyte' class is unused i.e. set to \".\""
  :type 'regexp
  :group 'convenience))

(when (featurep 'xemacs)
  (defcustom lower "[a-zäöüß]"
  "Rexexp to specify the character class 
Change it, if you want to pick strings differently.
XEmacs-users: `unibyte' and  `multibyte' class is unused i.e. set to \".\""
  :type 'regexp
  :group 'convenience))

(when (featurep 'xemacs)
  (defcustom multibyte "[.]"
  "Rexexp to specify the character class 
Change it, if you want to pick strings differently.
XEmacs-users: `unibyte' and  `multibyte' class is unused i.e. set to \".\""
  :type 'regexp
  :group 'convenience))

(when (featurep 'xemacs)
  (defcustom nonascii "[^\040-\177]"
  "Rexexp to specify the character class 
Change it, if you want to pick strings differently.
XEmacs-users: `unibyte' and  `multibyte' class is unused i.e. set to \".\""
  :type 'regexp
  :group 'convenience))

(when (featurep 'xemacs)
  (defcustom print "[\041-\177\241-\377]"
  "Rexexp to specify the character class 
Change it, if you want to pick strings differently.
XEmacs-users: `unibyte' and  `multibyte' class is unused i.e. set to \".\""
  :type 'regexp
  :group 'convenience))

(when (featurep 'xemacs)
  (defcustom punct "[.,-_:;?!]"
  "Rexexp to specify the character class 
Change it, if you want to pick strings differently.
XEmacs-users: `unibyte' and  `multibyte' class is unused i.e. set to \".\""
  :type 'regexp
  :group 'convenience))

(when (featurep 'xemacs)
  (defcustom space "[ \t]"
  "Rexexp to specify the character class 
Change it, if you want to pick strings differently.
XEmacs-users: `unibyte' and  `multibyte' class is unused i.e. set to \".\""
  :type 'regexp
  :group 'convenience))

(when (featurep 'xemacs)
  (defcustom unibyte "[.]"
  "Rexexp to specify the character class 
Change it, if you want to pick strings differently.
XEmacs-users: `unibyte' and  `multibyte' class is unused i.e. set to \".\""
  :type 'regexp
  :group 'convenience))

(when (featurep 'xemacs)
  (defcustom upper "[A-ZÄÖÜ]"
  "Rexexp to specify the character class 
Change it, if you want to pick strings differently.
XEmacs-users: `unibyte' and  `multibyte' class is unused i.e. set to \".\""
  :type 'regexp
  :group 'convenience))

(when (featurep 'xemacs)
  (defcustom xdigit "[0-9.,]"
    "Rexexp to specify the character class 
Change it, if you want to pick strings differently.
XEmacs-users: `unibyte' and `multibyte' class is unused i.e. set to \".\""
    :type 'regexp
    :group 'convenience))


;; Ascii

(put 'ascii 'beginning-op-at
     (lambda ()
       (if (featurep 'xemacs)
          (progn
	   (while 
	       (and (not (bobp))(looking-at ascii))
	     (forward-char -1))
           (unless (bobp) (forward-char 1))) 
       ;; (when
         ;; (looking-at "[[:ascii:]]")
       (skip-chars-backward "[:ascii:]"))))

(put 'ascii 'end-op-at
     (lambda ()
       (if (featurep 'xemacs)
           (progn (forward-char 1)
	     (while 
	       (looking-at ascii)
	     (forward-char 1)))
	 (skip-chars-forward "[:ascii:]"))))

;; Alnum

(put 'alnum 'beginning-op-at
     (lambda ()
       (if (featurep 'xemacs)
          (progn
	   (while 
	       (and (not (bobp))(looking-at alnum))
	     (forward-char -1))
           (unless (bobp) (forward-char 1))) 
       ;; (when
         ;; (looking-at "[[:alnum:]]")
       (skip-chars-backward "[:alnum:]"))))

(put 'alnum 'end-op-at
     (lambda ()
       (if (featurep 'xemacs)
           (progn (forward-char 1)
	     (while 
	       (looking-at alnum)
	     (forward-char 1)))
	 (skip-chars-forward "[:alnum:]"))))

;; Alpha

(put 'alpha 'beginning-op-at
     (lambda ()
       (if (featurep 'xemacs)
          (progn
	   (while 
	       (and (not (bobp))(looking-at alpha))
	     (forward-char -1))
           (unless (bobp) (forward-char 1))) 
       ;; (when
         ;; (looking-at "[[:alpha:]]")
       (skip-chars-backward "[:alpha:]"))))

(put 'alpha 'end-op-at
     (lambda ()
       (if (featurep 'xemacs)
           (progn (forward-char 1)
	     (while 
	       (looking-at alpha)
	     (forward-char 1)))
	 (skip-chars-forward "[:alpha:]"))))

;; Blank

(put 'blank 'beginning-op-at
     (lambda ()
       (if (featurep 'xemacs)
          (progn
	   (while 
	       (and (not (bobp))(looking-at blank))
	     (forward-char -1))
           (unless (bobp) (forward-char 1))) 
       ;; (when
         ;; (looking-at "[[:blank:]]")
       (skip-chars-backward "[:blank:]"))))

(put 'blank 'end-op-at
     (lambda ()
       (if (featurep 'xemacs)
           (progn (forward-char 1)
	     (while 
	       (looking-at blank)
	     (forward-char 1)))
	 (skip-chars-forward "[:blank:]"))))

;; Cntrl

(put 'cntrl 'beginning-op-at
     (lambda ()
       (if (featurep 'xemacs)
          (progn
	   (while 
	       (and (not (bobp))(looking-at cntrl))
	     (forward-char -1))
           (unless (bobp) (forward-char 1))) 
       ;; (when
         ;; (looking-at "[[:cntrl:]]")
       (skip-chars-backward "[:cntrl:]"))))

(put 'cntrl 'end-op-at
     (lambda ()
       (if (featurep 'xemacs)
           (progn (forward-char 1)
	     (while 
	       (looking-at cntrl)
	     (forward-char 1)))
	 (skip-chars-forward "[:cntrl:]"))))

;; Digit

(put 'digit 'beginning-op-at
     (lambda ()
       (if (featurep 'xemacs)
          (progn
	   (while 
	       (and (not (bobp))(looking-at digit))
	     (forward-char -1))
           (unless (bobp) (forward-char 1))) 
       ;; (when
         ;; (looking-at "[[:digit:]]")
       (skip-chars-backward "[:digit:]"))))

(put 'digit 'end-op-at
     (lambda ()
       (if (featurep 'xemacs)
           (progn (forward-char 1)
	     (while 
	       (looking-at digit)
	     (forward-char 1)))
	 (skip-chars-forward "[:digit:]"))))

;; Graph

(put 'graph 'beginning-op-at
     (lambda ()
       (if (featurep 'xemacs)
          (progn
	   (while 
	       (and (not (bobp))(looking-at graph))
	     (forward-char -1))
           (unless (bobp) (forward-char 1))) 
       ;; (when
         ;; (looking-at "[[:graph:]]")
       (skip-chars-backward "[:graph:]"))))

(put 'graph 'end-op-at
     (lambda ()
       (if (featurep 'xemacs)
           (progn (forward-char 1)
	     (while 
	       (looking-at graph)
	     (forward-char 1)))
	 (skip-chars-forward "[:graph:]"))))

;; Lower

(put 'lower 'beginning-op-at
     (lambda ()
       (if (featurep 'xemacs)
          (progn
	   (while 
	       (and (not (bobp))(looking-at lower))
	     (forward-char -1))
           (unless (bobp) (forward-char 1))) 
       ;; (when
         ;; (looking-at "[[:lower:]]")
       (skip-chars-backward "[:lower:]"))))

(put 'lower 'end-op-at
     (lambda ()
       (if (featurep 'xemacs)
           (progn (forward-char 1)
	     (while 
	       (looking-at lower)
	     (forward-char 1)))
	 (skip-chars-forward "[:lower:]"))))

;; Multibyte

(put 'multibyte 'beginning-op-at
     (lambda ()
       (if (featurep 'xemacs)
          (progn
	   (while 
	       (and (not (bobp))(looking-at multibyte))
	     (forward-char -1))
           (unless (bobp) (forward-char 1))) 
       ;; (when
         ;; (looking-at "[[:multibyte:]]")
       (skip-chars-backward "[:multibyte:]"))))

(put 'multibyte 'end-op-at
     (lambda ()
       (if (featurep 'xemacs)
           (progn (forward-char 1)
	     (while 
	       (looking-at multibyte)
	     (forward-char 1)))
	 (skip-chars-forward "[:multibyte:]"))))

;; Nonascii

(put 'nonascii 'beginning-op-at
     (lambda ()
       (if (featurep 'xemacs)
          (progn
	   (while 
	       (and (not (bobp))(looking-at nonascii))
	     (forward-char -1))
           (unless (bobp) (forward-char 1))) 
       ;; (when
         ;; (looking-at "[[:nonascii:]]")
       (skip-chars-backward "[:nonascii:]"))))

(put 'nonascii 'end-op-at
     (lambda ()
       (if (featurep 'xemacs)
           (progn (forward-char 1)
	     (while 
	       (looking-at nonascii)
	     (forward-char 1)))
	 (skip-chars-forward "[:nonascii:]"))))

;; Print

(put 'print 'beginning-op-at
     (lambda ()
       (if (featurep 'xemacs)
          (progn
	   (while 
	       (and (not (bobp))(looking-at print))
	     (forward-char -1))
           (unless (bobp) (forward-char 1))) 
       ;; (when
         ;; (looking-at "[[:print:]]")
       (skip-chars-backward "[:print:]"))))

(put 'print 'end-op-at
     (lambda ()
       (if (featurep 'xemacs)
           (progn (forward-char 1)
	     (while 
	       (looking-at print)
	     (forward-char 1)))
	 (skip-chars-forward "[:print:]"))))

;; Punct

(put 'punct 'beginning-op-at
     (lambda ()
       (if (featurep 'xemacs)
          (progn
	   (while 
	       (and (not (bobp))(looking-at punct))
	     (forward-char -1))
           (unless (bobp) (forward-char 1))) 
       ;; (when
         ;; (looking-at "[[:punct:]]")
       (skip-chars-backward "[:punct:]"))))

(put 'punct 'end-op-at
     (lambda ()
       (if (featurep 'xemacs)
           (progn (forward-char 1)
	     (while 
	       (looking-at punct)
	     (forward-char 1)))
	 (skip-chars-forward "[:punct:]"))))

;; Space

(put 'space 'beginning-op-at
     (lambda ()
       (if (featurep 'xemacs)
          (progn
	   (while 
	       (and (not (bobp))(looking-at space))
	     (forward-char -1))
           (unless (bobp) (forward-char 1))) 
       ;; (when
         ;; (looking-at "[[:space:]]")
       (skip-chars-backward "[:space:]"))))

(put 'space 'end-op-at
     (lambda ()
       (if (featurep 'xemacs)
           (progn (forward-char 1)
	     (while 
	       (looking-at space)
	     (forward-char 1)))
	 (skip-chars-forward "[:space:]"))))

;; Unibyte

(put 'unibyte 'beginning-op-at
     (lambda ()
       (if (featurep 'xemacs)
          (progn
	   (while 
	       (and (not (bobp))(looking-at unibyte))
	     (forward-char -1))
           (unless (bobp) (forward-char 1))) 
       ;; (when
         ;; (looking-at "[[:unibyte:]]")
       (skip-chars-backward "[:unibyte:]"))))

(put 'unibyte 'end-op-at
     (lambda ()
       (if (featurep 'xemacs)
           (progn (forward-char 1)
	     (while 
	       (looking-at unibyte)
	     (forward-char 1)))
	 (skip-chars-forward "[:unibyte:]"))))

;; Upper

(put 'upper 'beginning-op-at
     (lambda ()
       (if (featurep 'xemacs)
          (progn
	   (while 
	       (and (not (bobp))(looking-at upper))
	     (forward-char -1))
           (unless (bobp) (forward-char 1))) 
       ;; (when
         ;; (looking-at "[[:upper:]]")
       (skip-chars-backward "[:upper:]"))))

(put 'upper 'end-op-at
     (lambda ()
       (if (featurep 'xemacs)
           (progn (forward-char 1)
	     (while 
	       (looking-at upper)
	     (forward-char 1)))
	 (skip-chars-forward "[:upper:]"))))

;; Xdigit

(put 'xdigit 'beginning-op-at
     (lambda ()
       (if (featurep 'xemacs)
          (progn
	   (while 
	       (and (not (bobp))(looking-at xdigit))
	     (forward-char -1))
           (unless (bobp) (forward-char 1))) 
       ;; (when
         ;; (looking-at "[[:xdigit:]]")
       (skip-chars-backward "[:xdigit:]"))))

(put 'xdigit 'end-op-at
     (lambda ()
       (if (featurep 'xemacs)
           (progn (forward-char 1)
	     (while 
	       (looking-at xdigit)
	     (forward-char 1)))
	 (skip-chars-forward "[:xdigit:]"))))

(setq thingatpt-delimited-list (list
                                'angled-no-nest
                                'angled-greater
                                'angled-lower
                                'backslashed
                                'braced
                                'bracketed
                                'commented
                                'delimited
                                'dollared
                                'doublequoted
                                'parentized
                                'singlequoted
                                'slashed))

(defvar thingatpt-listed nil
  "Forms listed with thingatpt-list")

(setq th-classes (list "alnum" "alpha" "ascii" "blank" "cntrl" "digit" "graph" "lower" "multibyte" "nonascii" "print" "punct" "space" "unibyte" "upper" "xdigit"))

(setq thingatpt-listed '(alnum alpha angled-no-nest angled-greater angled-lower ascii backslashed blank braced bracketed cntrl comment csv date defun delimited digit dollared doublequoted email filename float function graph ip isbn line list lower ml-text multibyte nonascii number page paragraph paren parentized phone print punct region sentence sexp singlequoted slashed space string sh-struct symbol td unibyte upper url whitespace word xdigit ))

(defun thingatpt-listed ()
  
  (interactive "*")
  (message "%s" thingatpt-listed))


;; opposite proceeding: from GNU Emacs subr.el
(when (featurep 'xemacs) 
  (defun looking-back (regexp &optional limit greedy)
    "Return non-nil if text before point matches regular expression REGEXP.
    Like `looking-at' except matches before point, and is slower.
    LIMIT if non-nil speeds up the search by specifying a minimum
    starting position, to avoid checking matches that would start
    before LIMIT.
    If GREEDY is non-nil, extend the match backwards as far as possible,
    stopping when a single additional previous character cannot be part
    of a match for REGEXP."
    (let ((start (point))
	  (pos
	   (save-excursion
	     (and (re-search-backward (concat "\\(?:" regexp "\\)\\=") limit t)
		  (point)))))
      (if (and greedy pos)
	  (save-restriction
	    (narrow-to-region (point-min) start)
	    (while (and (> pos (point-min))
			(save-excursion
			  (goto-char pos)
			  (backward-char 1)
			  (looking-at (concat "\\(?:"  regexp "\\)\\'"))))
	      (setq pos (1- pos)))
	    (save-excursion
	      (goto-char pos)
	      (looking-at (concat "\\(?:"  regexp "\\)\\'")))))
    (not (null pos)))))


(defun face-atpt () 
  "Return face at point "
  (interactive)
  (let ((faceat (get-char-property (point) 'face)))
    (when (interactive-p) (message "%s" faceat))
    faceat))

(defun customize-face-atpt ()
  "Make customization buffer for face at point. "
  (interactive)
  (customize-variable (cadr (text-properties-at (point) (current-buffer)))))

(defvar corresponding-angle-atpt nil
  "For intern use with `angled'. ")

;; Angled

(put 'angled-no-nest 'beginning-op-at
     (lambda ()
       (if (looking-at "[<>]")
           (point)
         (re-search-backward  "[<>]" nil t 1))))


(put 'angled-no-nest 'end-op-at
     (lambda ()
       (search-forward   (cond ((equal (char-after) ?>) "<")
                               ((equal (char-after) ?<) ">")

                                ;; (equal (char-after) 62)
                               ;; ((equal (char-after) 60) ">")
                               (t nil))
               nil t 1)))

(defalias 'angled-atpt 'angled-no-nest-atpt)
(defalias 'bounds-of-angled-atpt 'bounds-of-angled-no-nest-atpt)

;; Angled-greater

(put 'angled-greater 'beginning-op-at
     (lambda ()
       (if (looking-at "<")
           (point)
         (beginning-of-form-base "<" ">" nil nil nil nil t))))


(put 'angled-greater 'end-op-at
     (lambda ()
       (end-of-form-base "<" ">" nil nil nil nil nil t)))

;; Angled-lower

(put 'angled-lower 'beginning-op-at
     (lambda ()
       (if (looking-at ">")
           (point)
         (beginning-of-form-base ">" "<" nil nil nil nil t))))


(put 'angled-lower 'end-op-at
     (lambda ()
       (end-of-form-base ">" "<" nil nil nil nil nil t)))

;; Backslashed

(put 'backslashed 'beginning-op-at
     (lambda ()
       (if (looking-at "\\\\")
	   (point)
	 (skip-chars-backward "^\\\\")
         (forward-char -1))))

(put 'backslashed 'end-op-at
     (lambda ()
       (forward-char 1) 
       (skip-chars-forward "^\\\\")
       (forward-char 1)))

;; Braced

(put 'braced 'beginning-op-at
     (lambda ()
       (if (looking-at "{")
           (point)
         (beginning-of-form-base "{" "}" nil nil nil nil t))))

(put 'braced 'end-op-at
     (lambda ()
       (end-of-form-base "{" "}" nil nil nil nil nil t)))

;; Bracketed


(put 'bracketed 'beginning-op-at
     (lambda ()
       (if (looking-at "\\[")
                       (point)
                       (beginning-of-form-base "[" "]" nil nil nil nil t))))

(put 'bracketed 'end-op-at
     (lambda ()
       (end-of-form-base "[" "]" nil nil nil nil nil t)))

;; Comment

;; (defvar comment-atpt-pos nil
;;   "For intern use only")

;; (defvar comment-found-by-pps nil
   ;; "For intern use only")

(put 'comment 'beginning-op-at
     (lambda ()
       (if (string= "" comment-end)
           (beginning-of-form-base comment-start comment-end (line-beginning-position) t nil t)
         (beginning-of-form-base comment-start comment-end nil t nil t))))

(put 'comment 'end-op-at
     (lambda ()
       (if (string= "" comment-end)
           (when (looking-at comment-start)
             (end-of-line)) 
         (end-of-form-base comment-start comment-end nil t nil t))))


;; (put 'comment 'beginning-op-at
;;      (lambda ()
;;        (setq comment-found-by-pps nil)
;;        (if
;;            (nth 4 (parse-partial-sexp (point-min) (point)))
;;            (progn
;;              (while (nth 4 (parse-partial-sexp (point-min) (point)))
;;                (forward-char -1))
;;              (while (eq (syntax-class (syntax-after (point)))
;;                         (syntax-class (syntax-after (1- (point)))))
;;                (forward-char -1))
;;              (setq comment-found-by-pps t))
;;          (unless (string= "" comment-end)
;;            (beginning-of-form-baseth comment-start comment-end nil t nil t t)))))
;; 
;; (put 'comment 'end-op-at
;;      (lambda ()
;;        (if comment-found-by-pps
;;            (progn 
;;              (forward-char (+ (length comment-start) (length comment-padding)))
;;              (while (and (nth 4 (parse-partial-sexp (point-min) (point)))(not (eobp))) 
;;                (forward-char 1))) 
;;          (unless (string= "" comment-end)
;;            (end-of-form-base comment-start comment-end nil t nil t t)))))

;;; CSV

;; Inspired by
;;; csv-mode.el --- major mode for editing comma-separated value files
;; Author: Francis J. Wright <F.J.Wright at qmul.ac.uk>
;; URL: http://centaur.maths.qmul.ac.uk/Emacs/


(defcustom csv-separator-atpt ";"
  "Char to distinguish datasets in a `comma`-separated row" 
  :type 'string
  :group 'convenience)

;; (when (boundp 'csv-separators)
;; (setq separator-atpt csv-separators))

(put 'csv 'beginning-op-at
     (lambda ()
       (search-backward (concat "^" csv-separator-atpt) nil nil 1)))

(put 'csv 'end-op-at
     (lambda ()
       (skip-chars-forward (concat "^" csv-separator-atpt))))

;; DATE

(put 'date 'beginning-op-at
     (lambda ()
       ;; provide for the case, we are over a
       ;; string-delimiter as `"'
       (when
	   (and (not (eq 32 (char-after))) 
		(or (bobp)
		    (eq 32 (char-before))))
	 (forward-char 1)
	 ;; as the bounds-function checks position, correct it
	 ;; (setq th-orig 1)
	 ) 
       (skip-chars-backward "[0-9] .-")
       (skip-chars-forward " ") 
))

(put 'date 'end-op-at
     (lambda ()
              (skip-chars-forward "[0-9] .-")
	      (skip-chars-backward " ")))

;; Defun

(put 'defun 'beginning-op-at (lambda (&optional arg) (beginning-of-defun (or arg 1))))

(put 'defun 'end-op-at (lambda (&optional arg)(end-of-defun (or arg 1))))

;; Delimited


(put 'delimited 'beginning-op-at
     (lambda ()
       (unless delimiters-atpt
	   (setq delimiters-atpt "\"'#$/=?!:;"))
       (if
	   (looking-at (concat "[^" delimiters-atpt "]"))
	   (progn 
	     (re-search-backward (concat "[" delimiters-atpt "]") nil nil 1)
	     (setq delimiter-zeichen-atpt (char-after)))
	 (setq delimiter-zeichen-atpt nil)))) 

(put 'delimited 'end-op-at
     (lambda ()
       (forward-char 1)
     (search-forward (char-to-string delimiter-zeichen-atpt) nil nil 1)))

;; (put 'delimited 'beginning-op-at
;;      (lambda ()
;;        (when
;; 	   (looking-at (concat "[^" delimiters-atpt "]"))
;; 	 (re-search-backward (concat "[" delimiters-atpt "]") nil nil 1)
;; 	 (setq delimiter-zeichen-atpt (char-after))  
;; 	   (when with-delimiters-atpt
;; 	     (forward-char -1))
;; t)))
;; 
;; (put 'delimited 'end-op-at
;;      (lambda ()
;;        (when with-delimiters-atpt
;; 	 (forward-char 2))
;;        (re-search-forward (char-to-string delimiter-zeichen-atpt) nil nil 1)))


(defvar delimiter-zeichen-atpt nil
  "Delimiter char found at place, search it backward then")

;; (defvar delimiters-atpt nil
;; "Accepted delimiters chars")

(defcustom with-delimiters-atpt  t 
 "If containing delimiters are copied with string. 
Default is t" 

:type 'boolean
:group 'convenience
)

(defcustom delimiters-atpt "\"'#$/=?!:;"
  "Specify the accepted delimiter chars."
  :type 'string
  :group 'convenience)

(defcustom th-beg-delimiter "{>[(/"
  "Specify the delimiter char."
  :type 'string
  :group 'convenience)
				
(defcustom th-end-delimiter "}]<)/"
  "Specify the delimiter char."
  :type 'string
  :group 'convenience)

;; (defun set-delimiter-atpt (&optional delimiters-atpt)
;; "Enables interactive setting of delimiters. "
;; (if delimiters-at
;; (setq th-beg-delimiter (car delimiters-atpt) th-end-delimiter (cdr delimiters-atpt))
;; (setq th-beg-delimiter
;; (read-from-minibuffer "th-beg-delimiter: " th-beg-delimiter))
;; (setq th-end-delimiter
;; (read-from-minibuffer "th-end-delimiter: " th-end-delimiter))))

;; Doublequoted


(put 'doublequoted 'beginning-op-at
     (lambda ()
       (when (eq 34 (nth 3 (parse-partial-sexp (point-min) (point))))
              (search-backward "\"" nil nil 1))))

(put 'doublequoted 'end-op-at
     (lambda ()
       (when (looking-at "\"") 
	 (search-forward "\"" nil nil 2))))

;; Email

(put 'email 'beginning-op-at
     (lambda ()
       (when
	   (looking-at "[^ \t]")
	 (re-search-backward "[,;][[:graph:]]\\|<[[:graph:]]\\|^[[:graph:]]\\|[^[:graph:]][[:graph:]]" (line-beginning-position) t 1)(when (looking-at "[[:space:];,<]") (forward-char 1))))) 

(put 'email 'end-op-at (lambda () (when (looking-at "[ <]\\{0,1\\}\\([\041-\132\136-\176]+@[\041-\132\136-\176]+\\)[;,> \t\n]*")
				 (goto-char (match-end 1))
				 (skip-chars-backward "[[:punct:]]")))) 

			   ;; (put 'email 'end-op-at (lambda () (when (looking-at "[ <]\\{0,1\\}\\([[:graph:]]+@[[:graph:]]+\\)[;,> \t\n]*")
			   ;; 			   (goto-char (match-end 1))
			   ;; 			   (skip-chars-backward "[[:punct:]]"))))

;; Filename

; (setq thingatpt-file-name-chars "-~/A-Za-z0-9ÄÖÜäöüß_.$?={}#%,: ") 
(if (featurep 'xemacs)
    (defcustom thingatpt-file-name-chars "-~/A-Za-z0-9ÄÖÜäöüß_.$?={}#%,: " 
      "Characters forseen in filenames. "
      :type 'string
      :group 'convenience)
  (defcustom thingatpt-file-name-chars "-~/[:alnum:]_.$?={}#%,:" 
    "Characters forseen in filenames. "
    :type 'string
    :group 'convenience)) 

; (defcustom thingatpt-file-name-chars  "-~/[:alnum:]_.$?={}#%,:" 
; "Characters forseen in filenames. "
; :type 'string
; :group 'convenience)

(put 'filename 'beginning-op-at
     (lambda ()
       (unless (looking-back " \t\n\r\f")
	 (skip-chars-backward thingatpt-file-name-chars))))

(put 'filename 'end-op-at
     (lambda ()
       (re-search-forward (concat "\\=[" thingatpt-file-name-chars "]*")
			  nil t)
     (skip-chars-backward ": ")))

;; Floats

(put 'float 'beginning-op-at (lambda () (when (numberp (read (buffer-substring-no-properties (point) (1+ (point)))))
    (skip-chars-backward "[0-9].,"))))

(put 'float 'end-op-at (lambda () (skip-chars-forward "[0-9.,]")))

;; Function

(put 'function 'beginning-op-at
     (lambda ()
       (unless (or
		(and (eq major-mode 'sh-mode) (looking-at beg-sh-struct-atpt))
		(and (eq 4 (car (syntax-after (point)))) (eq 0 (car (parse-partial-sexp (point-min) (point))))))
	 (beginning-of-defun)))) 

(put 'function 'end-op-at
     (lambda ()
       (end-of-defun)
       (when (string= major-mode "emacs-lisp-mode")
	 (skip-chars-backward " \t\r\n\f")))) 

;; IP

(put 'ip 'beginning-op-at
     (lambda ()
       (unless (looking-at "\\s-")
	 (skip-chars-backward "[0-9]."))))

(put 'ip 'end-op-at
     (lambda ()
       (when (looking-at "[0-9]\\{1,3\\}.[0-9-]\\{1,3\\}.[0-9]\\{1,3\\}.[0-9]\\{1,3\\}")
	 (goto-char (match-end 0)))))

;; ISBN

(put 'isbn 'beginning-op-at
     (lambda ()
       (unless (looking-at "\\s-")
	 (skip-chars-backward "[0-9]-"))))

(put 'isbn 'end-op-at
     (lambda ()
       (when (looking-at "[0-9]\\{1,3\\}[0-9-]\\{7,12\\}[0-9X]\\{0,1\\}")
	 (goto-char (match-end 0)))))

;; Lines

(put 'line 'beginning-op-at (lambda () (beginning-of-line)))
(put 'line 'end-op-at (lambda () (end-of-line)))
   
;;  Lists  


(put 'list 'end-op-at (lambda () (forward-list 1)
)) 

(put 'list 'beginning-op-at
     (lambda ()
       (unless (looking-at "\\s(")
	 (if (nth 9 (parse-partial-sexp (defun-beginning-position) (point)))
	     (goto-char (car (last (nth 9 (parse-partial-sexp (defun-beginning-position) (point))))))
	   (backward-sexp)
	   (forward-char -1)))))

;; Ml-Text
;; Useful to extract texts between ml-tags

(put 'ml-text 'beginning-op-at
     (lambda ()
       (when
	   (looking-at "[^>]") 
	 (re-search-backward ">" nil nil 1)
         (forward-char 1)))) 

(put 'ml-text 'end-op-at (lambda () (re-search-forward "</" nil nil 1) (forward-char -2))) 

;; Number

(put 'number 'beginning-op-at
     (lambda ()
       (when
	   (numberp (string-to-number (buffer-substring-no-properties (point) (1+ (point)))))
	 (skip-chars-backward "[0-9]"))))

(put 'number 'end-op-at
 (lambda ()
   (skip-chars-forward "[0-9]")))


;; Page

(put 'page 'beginning-op-at
     (lambda ()
       (backward-page)))


(put 'page 'end-op-at
     (lambda ()
       (forward-page)))


;; Paragraph

(put 'paragraph 'beginning-op-at
     (lambda ()
       (backward-paragraph)))


(put 'paragraph 'end-op-at
     (lambda ()
       (forward-paragraph)))

;; Paren
(put 'paren 'beginning-op-at
     (lambda ()
       (cond
	((looking-at "\\s)")
	 (forward-char 1) (backward-list 1))
	(t (while
	       (and (not (looking-at "\\s(")) (not (bobp)))
	     (forward-char -1))))))

(put 'paren 'end-op-at
     (lambda ()
       (forward-list 1)))

;; Parentized

(put 'parentized 'beginning-op-at
     (lambda ()
       (if (looking-at "(")
           (point)
         (beginning-of-form-base "(" ")" nil nil nil nil t))))
     
(put 'parentized 'end-op-at
     (lambda ()
       (if (looking-at "(")
           (forward-list)
         (skip-chars-forward "^)")
         (end-of-form-base "(" ")" nil nil nil nil nil t))))

;; (put 'parentized 'beginning-op-at
;; (lambda ()
;; (if (looking-at "(")
;; (point)
;; (beginning-of-form-base "(" ")" nil nil nil nil t))))
;; 
;; (put 'parentized 'end-op-at
;; (lambda ()
;; (if (looking-at "(")
;; (forward-list) 
;; (end-of-form-base "(" ")" nil nil nil nil nil t))))

;; Phone

(put 'phone 'beginning-op-at
     (lambda ()
       (when
	   (and (looking-at "[0-9 \t.()-]")
		(not (eq (char-before) ?+)))
       (re-search-backward "[^0-9 \t.()-][0-9 ()\t-]+" (line-beginning-position) nil 1) (forward-char 1)))) 

(put 'phone 'end-op-at
     (lambda ()
       (when
	   (looking-at "[0-9;, \t()-]")
	 (re-search-forward "[0-9 \t.()-]+[^0-9 \t-]" (1+ (line-end-position)) nil 1) (forward-char -1)))) 

;; Region

(defvar region-end-atpt nil)

(put 'region 'end-op-at
     (lambda ()
       (goto-char region-end-atpt)))

(put 'region 'beginning-op-at
     (lambda ()
       (setq region-end-atpt (region-end))
       (goto-char (region-beginning))))


;; Sentence

(put 'sentence 'beginning-op-at
     (lambda ()
       (backward-sentence)))


(put 'sentence 'end-op-at
     (lambda ()
       (forward-sentence)))


;; Sexp

(put 'sexp 'beginning-op-at
     (lambda ()
       (backward-sexp)))


(put 'sexp 'end-op-at
     (lambda ()
       (forward-sexp)))

;; Singlequoted

(put 'singlequoted 'beginning-op-at
     (lambda ()
       (when
	   (looking-at "[^']")
	 (search-backward  "'" nil nil 1))))


(put 'singlequoted 'end-op-at
     (lambda ()
       (search-forward "'" nil nil 2)))

;; Slashed

(put 'slashed 'beginning-op-at
     (lambda ()
       (if (looking-at "/")
	   (point)
	 (search-backward "/" nil nil 1))))

(put 'slashed 'end-op-at
     (lambda ()
       (search-forward "/" nil nil 2)))

;; Strings

(defvar th-string-beginningp nil)

(put 'string 'beginning-op-at
     (lambda ()
       (setq th-string-beginningp nil)
       (let ((pos (with-syntax-table (standard-syntax-table) (nth 8 (parse-partial-sexp (defun-beginning-position) (point))))))
	 (when pos
	   (goto-char pos)
	   (setq th-string-beginningp t)))))

(put 'string 'end-op-at
     (lambda ()
       (when th-string-beginningp
       (let ((pos (progn (save-excursion (beginning-of-defun) (point))))) 
	 (forward-char 1) 
	 (while (not (eq (char-after) (nth 3 (with-syntax-table (standard-syntax-table) (parse-partial-sexp pos (point))))))
	   (forward-char 1)))
       (forward-char 1)))) 

;; Sh-struct
(put 'sh-struct 'beginning-op-at 'sh-beginning-of-form)
(put 'sh-struct 'end-op-at
     (lambda ()
       (sh-end-of-form)
       (forward-char 1))) 
	     
(defcustom beg-sh-struct-atpt "case\\|for\\|function\\|if\\|select\\|until\\|while"
  "Specify the sh-struct beginning."
  :type 'regexp
  :group 'convenience)

(defcustom end-sh-struct-atpt "done\\|esac\\|fi" 
  "Specify the sh-struct end."
  :type 'regexp
  :group 'convenience)

;; Symbol

(put 'symbol 'beginning-op-at
     (lambda ()
       (unless (looking-at "\\s-")
	 (skip-syntax-backward "w_"))))

(put 'symbol 'end-op-at
     (lambda ()
       (skip-syntax-forward "w_")))

;; Td

(put 'td 'beginning-op-at
     (lambda ()
       (if (looking-at "<td[ >]")
           (point)
         (beginning-of-form-base "<td[ >]" "</td>"))))

(put 'td 'end-op-at
     (lambda ()
       (end-of-form-base "<td[ >]" "</td>")))

;; Dollared

(put 'dollared 'beginning-op-at
     (lambda ()
       (if (looking-at "$")
	   (point)
	 (search-backward "$" nil nil 1))))

(put 'dollared 'end-op-at
     (lambda ()
       (forward-char 1) 
       (search-forward "$" nil nil 1)))

;; Url

;; derived from thingatpt.el
(defvar thing-at-point-utils-uri-schemes
  ;; Officials from http://www.iana.org/assignments/uri-schemes.html
  '("ftp" "http" "gopher" "mailto" "news" "nntp"
    "telnet" "wais" "file" "prospero" "z39" 
    "cid" "mid" "vemmi" "service" "imap" "nfs" "acap" "rtsp"
    "tip" "pop" "data" "dav" "opaquelocktoken" "sip" "tel" "fax"
    "modem" "ldap" "https" "soap" "urn" "go"
    "afs" "tn3270" "mailserver"
    "crid" "dict" "dns" "dtn" "h323" "im" "info" "ipp"
    "iris" "mtqp" "mupdate" "pres" "sips" "snmp" "tag"
    "tftp" "xmlrpc" "xmpp"
  ;; Compatibility
    "snews" "irc" "mms" "mmsh")
  "Uniform Resource Identifier (URI) Schemes.")

(put 'url 'beginning-op-at
     (lambda ()
       (if
           (or (eq 10 (char-after))
               (eq 32 (char-after))
               (eq 9 (char-after)))
           (error nil)
         (forward-word 1)
         (dolist (protokoll thing-at-point-utils-uri-schemes)
           (when (equal protokoll
                        (progn
                          (re-search-backward (concat "\\(" protokoll "\\)") (line-beginning-position) t 1)
                          (match-string-no-properties 0)))
             (goto-char (match-beginning 0)))))))

(put 'url 'end-op-at
     (lambda ()
       ;; (skip-chars-forward ":/?#[]@!$&()*+,;=[:alnum:]-._~")
       (skip-chars-forward url-at-point-chars)
	 (skip-chars-backward ":")))
			   
(defcustom url-at-point-chars ":/?#[]@!$&()*+,;=[:alnum:]-._~"
  "Chars which might compose a URL. " 
  :type 'string
  :group 'convenience)


;; Whitespace

(put 'whitespace 'beginning-op-at
     (lambda () (when (looking-at "[ \t]") (skip-chars-backward "[ \t\r\n\f[:blank:]]"))))

(put 'whitespace 'end-op-at (lambda () (skip-chars-forward "[ \t\r\n\f[:blank:]]")))

;; Word


(put 'word 'beginning-op-at
     (lambda () (unless (or (bobp)
                            (looking-at "\\W")
                            (looking-back "\\W"))
		  (forward-word -1)))) 

(put 'word 'end-op-at
     (lambda () (unless (looking-at "\\W")
                  (forward-word 1)))) 
(defvar thing-invisible nil
  "If thing is hided")
(make-variable-buffer-local 'thing-invisible) 

(defun list-of-things-atpt ()
  "Displays a list of objects which might be called as THING herewith
  Every THING is provided with a set of functions at, before and after point 
  - i.e. to call with ACTION-THING-at,  -bf, -af."
  (interactive)
  (message "%s" thingatpt-listed))

;;; highlight-thing stuff
;; Basicely inspired by `highlight-current-line.el'. Please
;; visit
;; http://www.emacswiki.org/emacs-en/HighlightCurrentLine
;; to get the credits.

(defgroup highlight-thing nil
  "Highlight thing where the cursor is in."
  :load 'highlight-thing
  :group 'faces)

(defface highlight-thing-face
  '((t (:background "grey50")))
    "Face used to highlight current line."
  :group 'highlight-thing)

(defvar highlight-thing-minor-mode nil
  "Non-nil if using highlight-thing mode as a minor mode.")
(make-variable-buffer-local 'highlight-thing-minor-mode)

(defvar highlight-thing-overlay
  (make-overlay 1 1)
  "Overlay for highlighting.")

(overlay-put highlight-thing-overlay
	     'face 'highlight-thing-face)

(defvar highlight-thing-beg-function nil)
(defvar highlight-thing-end-function nil)

;; (setq highlight-thing-beg-function 'word-beginning-position-atpt)
;; (setq highlight-thing-end-function 'word-end-position-atpt)

(defun highlight-thing-hook ()
  (let ((beg (funcall highlight-thing-beg-function))
        (end (funcall highlight-thing-end-function)))
    (highlight-thing-hook-intern beg end nil)))

(defun highlight-thing-hook-intern (beg end &optional arg)
  "Post-Command-Hook for highlighting."
  (condition-case ()
      (if highlight-thing-minor-mode
          (let ((arg (or arg 0)))
            (move-overlay highlight-thing-overlay
                          beg end (current-buffer))))
    (error nil)))

(defconst highlight-thing-no-color (if (boundp 'xemacs-logo)
                                              '[]
                                            nil)
  "'color' value that represents \"no color\".")

;; Compatibility code - Set highlight-foregroundcolor.
(defun highlight-thing-set-fg-color (color)
  "Set foregroundcolor for highlighting cursor-word to COLOR.
Key: \\[highlight-thing-set-fg-color]"
  (interactive "sForeground color (\"none\" means no color): ")
  (if (equal "none" color)
      (setq color highlight-thing-no-color))
  (set-face-foreground 'highlight-thing-face color))

;; Compatibility code - Set highlight-backgroundcolor.
(defun highlight-thing-set-bg-color (color)
  "Set backgroundcolor for highlighting cursor-word to COLOR.
Key: \\[highlight-thing-set-bg-color]"
  (interactive "sBackground color (\"none\" means no color): ")
  (if (equal "none" color)
      (setq color highlight-thing-no-color))
  (set-face-background 'highlight-thing-face color))

;; Enable/Disable Highlighting
(defun highlight-thing (&optional on-off local)
  "Switch highlighting of cursor-word ON-OFF
If LOCAL is non-nil, do so locally for the current buffer only."
  (cond
   (on-off
    (when
        (featurep 'xemacs)
      (make-local-hook 'post-command-hook))
    (add-hook 'post-command-hook 'highlight-thing-hook nil local))
   (t
    (remove-hook 'post-command-hook 'highlight-thing-hook t)
    (delete-overlay highlight-thing-overlay))))

;;;###autoload
(defun highlight-thing-minor-mode (&optional arg)
  "Toggle highlight-thing minor mode.
With ARG, turn minor mode on if ARG is positive, off otherwise.
You can customize the face of the highlighted line and whether the entire
line is hightlighted by customizing the group highlight-thing."
  (setq highlight-thing-minor-mode
        (if (null arg)
            (not highlight-thing-minor-mode)
          (> (prefix-numeric-value arg) 0)))
  (if highlight-thing-minor-mode
      (highlight-thing t t)
    (highlight-thing nil t)))

(defun file-name-nondirectory-atpt (&optional arg) 
  (interactive "p")
  (let ((erg (file-name-nondirectory (filename-atpt))))
    (when arg
      (kill-new erg)
      (message "%s" erg))
    erg))

(defun th (thing &optional arg) 
  "Returns a buffer substring according to THING.
  THING may be a well known form as `symbol',
  `list', `sexp', `defun' or a newly defined THING.
  When mark-thingatpt is `t' - the default - a found THING 
  is set as current region, enabling further action on them.
  "
  (condition-case nil
      (let* ((arg arg)
              (thing-mark-region
               (and arg (or (eq 1 arg) (eq 4 (prefix-numeric-value arg)))))
              (bounds-point-at (th-bounds thing arg))
              (type 
               (cond ((and
                       (car bounds-point-at)(cdr bounds-point-at))
                      (buffer-substring-no-properties (car bounds-point-at) (cdr bounds-point-at)))
                     (t nil))))
             (when (and (stringp type) thing-mark-region)
               (kill-new type)
               (th-mark thing bounds-point-at)
               (message "   %s" type) (error nil))
             type)
    (error nil)))

(defun th-bounds (thing &optional arg)
  "Determine the start and end buffer locations for the THING at point.
  THING is a symbol which specifies the kind entity you want.
  Possibilities include `symbol', `list', `sexp', `defun', `filename', `url',
  `word', `sentence', `whitespace', `line', `page' and others.
  Call THING by his name, i.e. word-atpt etc., see `list-of-things-at' to see what's implemented"
  (condition-case nil
      (save-excursion
	(let* ((orig (point))
               (with-delimiters (cond ((not (memq thing thingatpt-delimited-list))
                                       t)
                                      ((and arg (eq 4 (prefix-numeric-value arg)))
                                       (not with-delimiters-atpt))
                                      (t with-delimiters-atpt)))
               (beg (progn
                      (funcall (get thing 'beginning-op-at))
                      (point)))
               (end (progn
                      (funcall (get thing 'end-op-at))
                      (point) )))
          (when (eq beg (point-min))
            ;; point may habe been between bob and function-beginning
            (setq beg (progn
                        (funcall (get thing 'beginning-op-at))
                        (point))))
	  (when (and (not (eq beg end))(if (region-exists-p)(<= orig end)(< orig end))(<= beg orig))
            (unless with-delimiters (setq beg (1+ beg)))
            (unless with-delimiters (setq end (1- end)))
            (when arg (message "%s" (cons beg end)))
	    (cons beg end))))
    (error nil)))

(defun th-beg (thing &optional arg)
  (condition-case nil 
      (let* ((bounds (th-bounds thing))
	     (beg (car bounds)))
	(when arg
	  (message "   %s " beg)) 
	beg)
    (error nil)))

(defun th-end (thing &optional arg)
  (condition-case nil 
      (let* ((bounds (th-bounds thing))
	     (end (cdr bounds)))
	(when arg
	  (message "   %s "  end)) 
	end)
    (error nil)))

(defun th-gotobeg (thing &optional arg)
  "Goto char beginning, core function "
  (condition-case nil 
      (let* ((bounds (th-bounds thing))
	     (beg (car bounds)))
	(when arg
	  (message "   %s " beg))
	(goto-char beg))
    (error nil)))

(defun th-gotoend (thing &optional arg)
  "Goto char end, core function "
  (condition-case nil 
      (let* ((bounds (th-bounds thing))
	     (end (cdr bounds)))
	(when arg
	  (message "   %s " end))
	(goto-char (1- end)))
    (error nil)))

(defun th-length (thing &optional arg)
  (condition-case nil 
      (let* ((bounds (th-bounds thing))
	     (length (format "%f" (abs (- (car bounds) (cdr bounds))))))
	(when arg
	  (message "   %s " length)) 
	length)
    (error nil)))

(defun th-count-base (cla elt &optional beg end ratio iact)
  (let ((beg
         (cond (beg beg)
               ((region-exists-p)
                (region-beginning))
               (t
                (funcall (intern-soft (concat (format "%s" elt) "-beginning-position-atpt")))
                    )))
	(end
         (cond (end (copy-marker end))
               ((region-exists-p)
                (copy-marker (region-end)))
               (t
                (condition-case nil (copy-marker (funcall (intern-soft (concat (format "%s" elt) "-end-position-atpt")))) (error nil))
                ))
         ))
    (th-count elt cla beg end ratio iact)))

(defun th-count (thing cla &optional beg end ratio iact)
  (condition-case nil
      (let* ((beg (or beg (car (th-bounds thing))))
              (end (or end (cdr (th-bounds thing))))
              (matchcount 0)
              len (erg 0) rest)
             (if (featurep 'xemacs)
                 (setq matchcount (string-to-number (string-strip (count-matches (eval cla)) nil "a-z ")))
               (setq matchcount (string-to-number (format "%f" (count-matches (concat "[[:" (format "%s" cla) ":]]") (or beg (point)) (or end (point)))))))
             (if ratio
                 (when (< 0 matchcount)
                   (setq len (string-to-number (format "%f" (- end beg))))
                   (setq erg (/ matchcount len))
                   (when iact (message "erg: %f" erg)) 
                   erg)
               (when (and iact)
                 (message "   %s " matchcount)) 
               matchcount))
    (error nil)))

(defun th-escape (thing)
  " "
  (regexp-quote (buffer-substring-no-properties (th-beg thing)(th-end thing))))

(defun th-copy (thing &optional arg)
  (condition-case nil 
      (let ((newcopy (th thing)))
	(when newcopy
	    (progn
	      (kill-new newcopy)
	      (if arg
		  (message "   %s" (car kill-ring))
		(car kill-ring)))))
    (error nil)))

(defun th-trim (&optional thing left right)
    (let* ((bounds (th-bounds thing))
      (left (when left (car bounds)))
      (right (when right (cdr bounds))))
      
  (save-excursion
    (cond ((and left right)
	   (goto-char right)
	   (delete-char -1)
	   (goto-char left)
	   (delete-char 1))
	  (right
	   (goto-char right)
	   (delete-char -1))
	  (left
	   (goto-char left)
	   (delete-char 1))))))

(defun th-separate (thing &optional arg) 
  " "
  (condition-case nil 
      (let* ((bounds (th-bounds thing))
	      (beg (car bounds))
	      (end (if (listp (cdr bounds))
		       (cadr bounds)
		     (cdr bounds)))
	      (oldbufsize (buffer-size)))
	     (goto-char beg)
	     (when (and beg end (<= (line-beginning-position) beg) (not (looking-back "^[ \t]+")))
	       (goto-char beg)
	       (newline-and-indent))
	     (save-excursion 
	       (when
		   (< oldbufsize (buffer-size)) 
		 (setq end (+ end (- (buffer-size) oldbufsize)))
		 (setq beg (+ beg (- (buffer-size) oldbufsize)))
		 (setq oldbufsize (buffer-size)))
	       (goto-char end)
	       (cond ((eobp)
		      (newline-and-indent))
		     ((looking-at "[\t\r\n\f ]*$")
		      nil)
		     (t (newline-and-indent)))
	       (list beg end)))
    (error nil)))

(defun th-comment (thing &optional arg) 
  " "
  (condition-case nil 
      (let* ((bounds (th-bounds thing))
	     (beg (car bounds))
	     (end (cdr bounds)))
	(if (and beg end)
	    (progn 
	      (goto-char beg)
              (if (functionp 'comment-indent-lor)
                  (comment-indent-lor t beg end)
                (comment-dwim nil))
              nil)))
        (error nil)))

(defun th-kill (thing &optional arg) 
  " "
  (condition-case nil 
      (let* ((bounds (th-bounds thing)) 
	     (beg (car bounds))
	     (end (cdr bounds)))
	(kill-region beg end))
    (error nil)))

(defun th-mark (thing &optional bounds) 
  " "
  (condition-case nil 
      (let* ((bounds (or bounds (th-bounds thing))))
	(goto-char (car bounds))
	(push-mark)
	(goto-char (cdr bounds))
	(exchange-point-and-mark))
    (error nil)))

;; uses sgml-tag from sgml-mode.el
(defun th-hide (thing &optional beg end) 
  "Toggle visibility of existing things in buffer or region. "
  (let ((beg (cond (beg beg)
		   ((region-active-p)
		    (region-beginning))
		   (t (point-min)))) 
	(end (cond (end (copy-marker end))
		   ((region-active-p)
		    (copy-marker (region-end)))
		   (t (copy-marker (point-max)))))
        (modified (buffer-modified-p))
	(inhibit-read-only t))
    (unwind-protect
	(save-excursion
	  (goto-char beg)
	  (if (setq thing-invisible (not thing-invisible))
              (while (th-forward thing)
                (let* ((bounds (th-bounds thing))
                        (beg (car bounds))
                        (end (cdr bounds)))
                       (put-text-property beg end 'category 'sgml-tag)))
            (remove-text-properties beg end '(category nil))))
      (restore-buffer-modified-p modified)))
  (message ""))

(defun th-highlight (thing &optional arg) 
  (setq highlight-thing-beg-function (intern-soft (concat (format "%s" thing) "-beginning-position-atpt")))
  (setq highlight-thing-end-function (intern-soft (concat (format "%s" thing) "-end-position-atpt")))
  (highlight-thing-minor-mode))

(defun th-forward (thing &optional arg) 
  "Returns beg and end of THING after point as a list. "
  (condition-case nil 
      (let* ((orig (point))
	     (arg (or arg 1))
             bounds ap ep)
        (if (< 0 arg)
            (while (< 0 arg)
              (forward-char 1)
              (while (and (not (eobp))(not (setq bounds (th-bounds thing))))
                (forward-char 1))
              (setq ap (car bounds))
              (setq ep (cdr bounds))
              (if (and ep (< orig (1- ep)))
                  (progn (goto-char (1- ep))
                         (setq arg (1- arg)))))
          (while
              (> 0 arg)
            (forward-char -1)
            (while (and (not (eobp))(not (setq bounds (th-bounds thing))))
              (forward-char -1))
            (setq ap (car bounds))
            (setq ep (cdr bounds))
            (setq arg (1+ arg)))
          (goto-char ap))
        (when arg (message "%s" bounds)) 
        bounds)
    (error nil)))

(defun th-backward (thing &optional arg) 
 "Returns beg and end of THING before point as a list. "
  (condition-case nil
      (th-forward thing (- (or arg 1)))
    (error nil)))

(defun th-after (thing &optional arg)
  " "
  (let* ((arg (or arg 1)) 
         (bounds-point-at (th-bounds thing))
         (end (cdr-safe bounds-point-at))
         bounds)
    (when end (goto-char end))
    (setq bounds (th-forward thing arg))
    (th thing arg)))

(defun th-bounds-after (thing &optional arg)
  " "
  (save-excursion 
    (let* ((arg (or arg 1)) 
            (bounds-point-at (th-bounds thing))
            (end (cdr-safe bounds-point-at))
            bounds)
           (when end (goto-char end))
           (setq bounds (th-forward thing arg))
           (when arg (message "%s" bounds)) bounds)))

(defun th-before (thing &optional arg)
  " "
  (let* ((arg (or arg -1)) 
         (bounds-point-at (th-bounds thing))
         (beg (car-safe bounds-point-at))
         bounds)
    (when beg (goto-char beg))
    (setq bounds (th-backward thing (abs arg)))
    (th thing arg)))

(defun th-bounds-before (thing &optional arg)
  " "
  (save-excursion 
    (let* ((arg (or arg -1)) 
            (bounds-point-at (th-bounds thing))
            (beg (car-safe bounds-point-at))
            bounds)
           (when beg (goto-char beg))
           (setq bounds (th-backward thing (abs arg)))
           (when arg (message "%s" bounds)) bounds)))

(defvar paired-start-pos nil)

(defun th-transpose (thing arg)
  (let* ((arg (if arg (prefix-numeric-value arg) 1))
         (pos1 (th-bounds-before thing arg))
         (pos2 (th-bounds-after thing arg))
         (a (car pos1))
         (b (copy-marker (cdr pos1)))
         (c (car pos2))
         (d (copy-marker (cdr pos2))))
    (transpose-regions a b c d)
    (goto-char d)))

;; credits to sort-subr, sort.el
;; (reverse nextrecfun endrecfun &optional startkeyfun endkeyfun predicate)

(defun th-sort (thing reverse beg end startkeyfun endkeyfun predicate)
      (save-excursion
	(save-restriction
	  (narrow-to-region beg end)
	  (goto-char (point-min))
	  (let ((reverse (or reverse nil))
		(startkeyfun (or startkeyfun nil))
		(endkeyfun (or endkeyfun nil))
		(predicate (or predicate nil))
		(this-beg (th-beg thing 0)))
            (while (not (or (eobp)(stringp (th thing))))
              (forward-char 1)) 
            (sort-subr reverse (function (lambda () (let ((pos (point))) (th-forward thing 1) (if (< pos (point)) (th-gotobeg thing 0) nil)))) (function (lambda () (th-gotoend thing 1)(forward-char 1))) startkeyfun endkeyfun predicate)))))

(defun th-delim (thing &optional arg beg-char end-char)
  "Process begin and end of region according to value of 
  `delim-action'
  If no region is active, process borders of THING-at-point
  according to value of delim-action-beginning- resp. -end-position
  Default is symbol-at.
  With \C-u or arg `escaped' to `t': insert escaped doublequotes"
  (interactive "*p")
  (condition-case nil 
      (save-restriction
	(let* ((pos (point)) 
	       (bounds (th-bounds thing)) 
	       (beg (copy-marker (car bounds)))
	       (end (copy-marker (cdr bounds)))
	       (beg-char (or beg-char th-beg-delimiter))
	       (end-char (or end-char th-end-delimiter)))
	  ;; (narrow-to-region beg end)
	  ;; (point-min-marker)
	  ;; (point-max-marker)
	  (goto-char beg)
	  (insert beg-char)
	  (goto-char end)
	  (insert end-char)
	  (goto-char (1+ pos))
	  (when arg (message "%s" bounds))))
    (error nil)))

;; (insert
;;;###autoload 
(defun th-base-copy-or (kind arg)
  " "
  (let ((expr (format "%s" kind))) 
    (if (< 0 arg)
        (if (region-exists-p)
            (funcall (or (intern-soft (concat (substring expr 0 (string-match "ed$" expr)) "-region-atpt"))
                         (intern-soft (concat (substring expr 0 (string-match "d$" expr)) "-region-atpt"))))
          (funcall (intern-soft (concat expr "-atpt")) (prefix-numeric-value arg)))
      (funcall (intern-soft (concat "kill-" expr "-atpt"))))))

(defun th-backslash (thing &optional arg) 
  " "
  (th-delim thing arg backlash-beg-char backlash-end-char))

(setq backlash-end-char ?\\)
(setq backlash-beg-char ?\\)

(defun th-brace (thing &optional arg)
  " "
  (th-delim thing arg "\{" "\}"))

(defun th-bracket (thing &optional arg)
  " "
  (th-delim thing arg ?\[ ?\]))

(defun th-greater-angle (thing &optional arg)
  " "
  (th-delim thing arg ?\< ?\>))

(defun th-lower-angle (thing &optional arg)
  " "
  (th-delim thing arg ?\> ?\<))

(defun th-dollar (thing &optional arg)
  " "
  (th-delim thing arg ?\$ ?\$))

(defun th-doublequote (thing &optional arg)
  " "
  (th-delim thing arg ?\" ?\"))

(defun th-singlequote (thing &optional arg)
  " "
  (th-delim thing arg ?\' ?\'))

(defun th-parentize (thing &optional arg)
  " "
  (th-delim thing arg ?\( ?\)))

(defun th-slash (thing &optional arg)
  " "
  (th-delim thing arg ?\/ ?\/))

(defun syntax-class-atpt () 
  "Return the syntax class part of the syntax at point. "
  (interactive)
  (let ((erg (logand (car (syntax-after (point))) 65535)))
    (when (interactive-p) (message "%s" erg)) erg))

(defun syntax-class-bfpt () 
  "Return the syntax class part of the syntax at point. "
  (interactive)
  (let ((erg (logand (car (syntax-after (1- (point)))) 65535)))
    (when (interactive-p) (message "%s" erg)) erg))

(defun syntax-atpt () 
  (interactive)
  (let ((stax (syntax-after (point))))
    (when (interactive-p)
      (message (format "%s" stax)))
    (car stax)))

(defun syntax-bfpt () 
  (interactive)
  (let ((stax (syntax-after (1- (point)))))
    (when (interactive-p)
      (message (format "%s" stax)))
    (car stax)))

;; Until then when `up-list' was called while in
;;  string, it failed with message: 
;; Scan error: ... "Unbalanced parentheses" 

(defun list-up-atpt (&optional arg)
  "Move forward out of one level of parentheses.
With ARG, do this that many times.
A negative argument means move backward but still to a less deep spot."
  (interactive "p")
  
  (let ((inc (if (> arg 0) 1 -1)))
    (while (/= arg 0)
      (goto-char
       (or
	(progn
	  (when (in-string-p)
	    (goto-char (or (nth 8 (parse-partial-sexp (defun-beginning-position) (point))) (point)))
	    (forward-sexp 1))
	  (scan-lists (point) inc 1)) (buffer-end arg)))
      (setq arg (- arg inc)))))

(defun delist-list-atpt () 
  "Delete delimiters of a list at point"
  (interactive)
  (let ((poss (list-bounds-atpt)))
    (narrow-to-region (car poss) (if (listp (cdr poss))
				     (cadr poss)
				   (cdr poss)))
    (point-min-marker)
    (point-max-marker) 
    (goto-char (point-min-marker))
    (delete-char 1)
    (goto-char (point-max-marker)) 
    (backward-delete-char 1)
    (widen)))


(provide 'thingatpt-utils-base)
;;; thingatpt-utils-base.el ends here
